/*
 * Copyright (C) 1998-2013 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file fade.c A volume fader and alarm clock for OSS. */

#include <regex.h>

#include "fade.cmdline.h"
#include "para.h"
#include "fd.h"
#include "string.h"
#include "mix.h"
#include "error.h"
#include "ggo.h"
#include "version.h"

INIT_FADE_ERRLISTS;
static struct fade_args_info conf;

enum mixer_id {MIXER_ENUM};
static char *mixer_name[] = {MIXER_NAMES};
DECLARE_MIXER_INITS;
static struct mixer supported_mixer[] = {MIXER_ARRAY};
#define FOR_EACH_MIXER(i) for ((i) = 0; (i) < NUM_SUPPORTED_MIXERS; (i)++)

static int loglevel;
__printf_2_3 void date_log(int ll, const char *fmt, ...)
{
	va_list argp;
	time_t t1;
	struct tm *tm;

	if (ll < loglevel)
		return;
	time(&t1);
	tm = localtime(&t1);
	printf("%d:%02d:%02d ", tm->tm_hour, tm->tm_min, tm->tm_sec);
	va_start(argp, fmt);
	vprintf(fmt, argp);
	va_end(argp);
}
__printf_2_3 void (*para_log)(int, const char*, ...) = date_log;

/* Fade to new volume in fade_time seconds. */
static int fade(struct mixer *m, struct mixer_handle *h, int new_vol, int fade_time)
{
	int vol, diff, incr, ret;
	unsigned secs;
	struct timespec ts;
	unsigned long long tmp, tmp2; /* Careful with that axe, Eugene! */

	if (fade_time <= 0)
		return m->set(h, new_vol);
	secs = fade_time;
	PARA_NOTICE_LOG("fading to %d in %d seconds\n", new_vol, secs);
	ret = m->get(h);
	if (ret < 0)
		goto out;
	vol = ret;
	diff = new_vol - vol;
	if (!diff) {
		sleep(secs);
		goto out;
	}
	incr = diff > 0? 1: -1;
	diff = diff > 0? diff: -diff;
	tmp = secs * 1000 / diff;
	tmp2 = tmp % 1000;
	while ((new_vol - vol) * incr > 0) {
		ts.tv_nsec = tmp2 * 1000000; /* really nec ?*/
		ts.tv_sec = tmp / 1000; /* really nec ?*/
		//printf("ts.tv_sec: %i\n", ts.tv_nsec);
		vol += incr;
		ret = m->set(h, vol);
		if (ret < 0)
			goto out;
		//printf("vol = %i\n", vol);
		nanosleep(&ts, NULL);
	}
out:
	return ret;
}

static void client_cmd(const char *cmd)
{
	int ret, status, fds[3] = {0, 0, 0};
	pid_t pid;
	char *cmdline = make_message(BINDIR "/para_client %s", cmd);

	PARA_NOTICE_LOG("%s\n", cmdline);
	ret = para_exec_cmdline_pid(&pid, cmdline, fds);
	free(cmdline);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		goto fail;
	}
	do
		pid = waitpid(pid, &status, 0);
	while (pid == -1 && errno == EINTR);
	if (pid < 0) {
		PARA_ERROR_LOG("%s\n", strerror(errno));
		goto fail;
	}
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		goto fail;
	return;
fail:
	PARA_EMERG_LOG("command \"%s\" failed\n", cmd);
	exit(EXIT_FAILURE);
}

static void change_afs_mode_and_play(char *afs_mode)
{
	char *cmd;

	client_cmd("stop");
	if (!afs_mode)
		return;
	cmd = make_message("select %s", afs_mode);
	client_cmd(cmd);
	free(cmd);
	client_cmd("play");
}

static int sweet_dreams(struct mixer *m, struct mixer_handle *h)
{
	time_t t1, wake_time_epoch;
	unsigned int delay;
	struct tm *tm;
	int ret, min = conf.wake_min_arg;
	char *fo_mood = conf.fo_mood_arg;
	char *fi_mood = conf.fi_mood_arg;
	char *sleep_mood = conf.sleep_mood_arg;
	int fit = conf.fi_time_arg;
	int fot = conf.fo_time_arg;
	int fiv = conf.fi_vol_arg;
	int fov = conf.fo_vol_arg;
	int iv = conf.ivol_arg;

	/* calculate wake time */
	time(&t1);
	if (conf.wake_hour_given) {
		int hour = conf.wake_hour_arg;
		tm = localtime(&t1);
		if (tm->tm_hour > hour || (tm->tm_hour == hour && tm->tm_min> min)) {
			t1 += 86400; /* wake time is tomorrow */
			tm = localtime(&t1);
		}
		tm->tm_hour = hour;
		tm->tm_min = min;
		tm->tm_sec = 0;
	} else {
		t1 += 9 * 60 * 60; /* nine hours from now */
		PARA_INFO_LOG("default wake time: %lu\n", (long unsigned)t1);
		tm = localtime(&t1);
	}
	wake_time_epoch = mktime(tm);
	PARA_INFO_LOG("waketime: %s", asctime(tm));
	client_cmd("stop");
	sleep(1);
	if (fot) {
		PARA_INFO_LOG("initial volume: %d\n", iv);
		ret = m->set(h, iv);
		if (ret < 0)
			return ret;
		change_afs_mode_and_play(fo_mood);
		ret = fade(m, h, fov, fot);
		if (ret < 0)
			return ret;
	} else {
		ret = m->set(h, fov);
		if (ret < 0)
			return ret;
	}
	if (conf.sleep_mood_given)
		change_afs_mode_and_play(sleep_mood);
	else
		client_cmd("stop");
	if (!fit)
		return 1;
	for (;;) {
		time(&t1);
		if (wake_time_epoch <= t1 + fit)
			break;
		delay = wake_time_epoch - t1 - fit;
		PARA_INFO_LOG("sleeping %u seconds (%u:%02u)\n",
			delay, delay / 3600,
			(delay % 3600) / 60);
		sleep(delay);
	}
	change_afs_mode_and_play(fi_mood);
	ret = fade(m, h, fiv, fit);
	PARA_INFO_LOG("fade complete, returning\n");
	return ret;
}

static int snooze(struct mixer *m, struct mixer_handle *h)
{
	int ret, val;

	if (conf.so_time_arg <= 0)
		return 1;
	ret = m->get(h);
	if (ret < 0)
		return ret;
	val = ret;
	if (val < conf.so_vol_arg)
		ret = m->set(h, conf.so_vol_arg);
	else
		ret = fade(m, h, conf.so_vol_arg, conf.so_time_arg);
	if (ret < 0)
		return ret;
	client_cmd("pause");
	PARA_NOTICE_LOG("%d seconds snooze time...\n", conf.snooze_time_arg);
	sleep(conf.snooze_time_arg);
	client_cmd("play");
	return fade(m, h, conf.si_vol_arg, conf.si_time_arg);
}

static int configfile_exists(void)
{
	static char *config_file;

	if (!conf.config_file_given) {
		char *home = para_homedir();
		free(config_file);
		config_file = make_message("%s/.paraslash/fade.conf", home);
		free(home);
		conf.config_file_arg = config_file;
	}
	return file_exists(conf.config_file_arg);
}

static void init_mixers(void)
{
	int i;

	FOR_EACH_MIXER(i) {
		struct mixer *m = &supported_mixer[i];
		PARA_DEBUG_LOG("initializing mixer API #%d (%s)\n",
			i, mixer_name[i]);
		m->init(m);
	}
}

static int set_channel(struct mixer *m, struct mixer_handle *h)
{
	char *channels;
	int ret;

	ret = m->set_channel(h, conf.mixer_channel_arg);
	if (ret >= 0) {
		PARA_NOTICE_LOG("using %s mixer channel\n",
			conf.mixer_channel_arg?  conf.mixer_channel_arg
				: "default");
		return ret;
	}
	channels = m->get_channels(h);
	printf("Available channels: %s\n", channels);
	free(channels);
	return ret;
}

static struct mixer *get_mixer_or_die(void)
{
	int i;

	if (!conf.mixer_api_given)
		i = DEFAULT_MIXER;
	else
		FOR_EACH_MIXER(i)
			if (!strcmp(mixer_name[i], conf.mixer_api_arg))
				break;
	if (i < NUM_SUPPORTED_MIXERS) {
		PARA_NOTICE_LOG("using %s mixer API\n", mixer_name[i]);
		return supported_mixer + i;
	}
	printf("available mixer APIs: ");
	FOR_EACH_MIXER(i) {
		int d = (i == DEFAULT_MIXER);
		printf("%s%s%s ", d? "[" : "", mixer_name[i], d? "]" : "");
	}
	printf("\n");
	exit(EXIT_FAILURE);
}

__noreturn static void print_help_and_die(void)
{
	struct ggo_help h = DEFINE_GGO_HELP(fade);
	bool d = conf.detailed_help_given;

	ggo_print_help(&h, d? GPH_STANDARD_FLAGS_DETAILED : GPH_STANDARD_FLAGS);
	exit(0);
}

int main(int argc, char *argv[])
{
	int ret;
	struct mixer *m;
	struct mixer_handle *h = NULL;

	fade_cmdline_parser(argc, argv, &conf);
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	version_handle_flag("fade", conf.version_given);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	ret = configfile_exists();
	if (!ret && conf.config_file_given) {
		PARA_EMERG_LOG("can not read config file %s\n",
			conf.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (ret) {
		struct fade_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 1
		};
		fade_cmdline_parser_config_file(conf.config_file_arg,
			&conf, &params);
		loglevel = get_loglevel_by_name(conf.loglevel_arg);
	}
	init_mixers();
	m = get_mixer_or_die();
	ret = m->open(conf.mixer_device_arg, &h);
	if (ret < 0)
		goto out;
	ret = set_channel(m, h);
	if (ret < 0)
		goto out;
	switch (conf.mode_arg) {
	case mode_arg_fade:
		ret = fade(m, h, conf.fade_vol_arg, conf.fade_time_arg);
		break;
	case mode_arg_snooze:
		ret = snooze(m, h);
		break;
	default: /* sleep mode */
		ret = sweet_dreams(m, h);
		break;
	}
out:
	m->close(&h);
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
